from datetime import datetime, timezone
import httplib2

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
import django.contrib.auth.password_validation as validators
from django.core import exceptions as core_exceptions
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions

from rest_framework_json_api import serializers
from rest_framework_json_api.relations import ResourceRelatedField

from oauth2client.client import \
    AccessTokenCredentials, \
    AccessTokenCredentialsError
from googleapiclient.discovery import build

import facebook

from jasonpi.models import Provider
from jasonpi.normalizers import facebook_profile, google_profile

User = get_user_model()


def assign(instance, profile):
    for key in profile:
        if hasattr(instance, key) and (
                getattr(instance, key) is None or getattr(instance, key) == ''
        ):
            setattr(instance, key, profile[key])


class GroupSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Group
        fields = ('url', 'name')


class ProviderSerializer(serializers.HyperlinkedModelSerializer):
    access_token = serializers.CharField(max_length=255, write_only=True)

    def __init__(self, *args, **kwargs):
        self.user = None
        if 'context' in kwargs and 'request' in kwargs['context']:
            user = kwargs['context']['request'].user
            if user.is_authenticated:
                self.user = user
        super(ProviderSerializer, self).__init__(*args, **kwargs)

    def validate_google(self, access_token, uid):
        try:
            credential = AccessTokenCredentials(
                access_token,
                'jasonpi/1.0'
            )
        except AccessTokenCredentialsError:
            raise exceptions.ValidationError(_('Invalid access token'))
        http = httplib2.Http()
        http = credential.authorize(http)
        service = build("people", "v1", http=http)
        user = service.people().get(
            resourceName='people/me',
            personFields='addresses,emailAddresses,names,genders,birthdays'
        ).execute()
        user_id = user['resourceName'].split('/')[1]
        if uid != user_id:
            raise exceptions.ValidationError(
                _('Google user id doesn\'t match'))
        return google_profile(user)

    def validate_facebook(self, access_token, uid):
        graph = facebook.GraphAPI(access_token=access_token)
        user = graph.get_object(
            id='me',
            fields='picture,first_name,last_name,name,birthday,gender,email',
        )
        if user['id'] != uid:
            raise exceptions.ValidationError(
                _('Facebook user id doesn\'t match'))
        return facebook_profile(user)

    def get_unique_together_validators(self):
        return []

    def validate(self, data):
        access_token = data.get('access_token')
        uid = data.get('uid')
        provider = data.get('provider')
        if provider == 'google':
            profile = self.validate_google(access_token, uid)
        elif provider == 'facebook':
            profile = self.validate_facebook(access_token, uid)
        else:
            raise exceptions.ValidationError(_('Provider not supported'))
        data['profile'] = profile
        if self.user is None:
            return super(ProviderSerializer, self).validate(data)
        try:
            pu = Provider.objects.get(
                uid=uid,
                provider=provider,
            )
            if pu.user != self.user:
                pass
                raise exceptions.ValidationError(
                    _('The social account you\'re to trying to use '
                      'is already linked to another user.')
                )
        except Provider.DoesNotExist:
            pass
        try:
            email_user = User.objects.get(email=profile['email'])
            if email_user != self.user:
                raise exceptions.ValidationError(
                    _('The email address for this social account is used by '
                      'another user.')
                )
        except User.DoesNotExist:
            pass
        return super(ProviderSerializer, self).validate(data)

    def save(self):
        current_user = self.user
        profile = self.validated_data['profile']
        uid = self.validated_data['uid']
        provider = self.validated_data['provider']

        try:
            up = Provider.objects.get(
                uid=uid,
                provider=provider
            )
            self.instance = up.user
            return
        except Provider.DoesNotExist:
            pass

        if current_user:
            user = current_user
        else:
            try:
                user = User.objects.get(email=profile['email'])
            except User.DoesNotExist:
                kwargs = init_kwargs(User, profile)
                user = User(**kwargs)
        assign(user, profile)
        user.save()
        Provider.objects.create(
            user=user,
            uid=uid,
            provider=provider
        )
        self.instance = user

    class Meta:
        model = Provider
        fields = ('url', 'uid', 'provider', 'access_token')


class UserSerializer(serializers.HyperlinkedModelSerializer):
    groups = ResourceRelatedField(
        queryset=Group.objects,
        many=True,
        related_link_view_name='user-groups-list',
        related_link_url_kwarg='user_pk',
        self_link_view_name='user-relationships',
        required=False,
    )

    providers = ResourceRelatedField(
        queryset=Provider.objects,
        many=True,
        related_link_view_name='user-providers-list',
        related_link_url_kwarg='user_pk',
        self_link_view_name='user-relationships',
        required=False,
    )

    included_serializers = {
        'groups': GroupSerializer,
        'providers': ProviderSerializer,
    }

    password = serializers.CharField(write_only=True)
    old_password = serializers.CharField(write_only=True, required=False)

    def validate(self, data):
        if self.instance:
            user = self.instance
        else:
            user = User(**data)

        password = data.get('password')
        if password is None:
            # if password is None it means it is not required
            return super(UserSerializer, self).validate(data)

        errors = dict()
        try:
            validators.validate_password(password=password, user=user)
        except core_exceptions.ValidationError as e:
            errors['password'] = list(e.messages)
        except Exception as e:
            raise e
        if self.instance and \
                not self.instance.check_password(data.get('old_password')):
            errors['old_password'] = _('Old password isn\'t valid.')

        if errors:
            raise serializers.ValidationError(errors)

        return super(UserSerializer, self).validate(data)

    def create(self, validated_data):
        user = User(**validated_data)
        user.last_login = datetime.now(timezone.utc)
        user.set_password(validated_data['password'])
        user.save()
        return user

    def update(self, instance, validated_data):
        instance = super(UserSerializer, self).update(instance, validated_data)
        password = validated_data.get('password')
        if password is not None:
            instance.set_password(password)
            instance.save()
        return instance

    def get_root_meta(self, resource, many):
        if hasattr(self, 'token'):
            return {
                'access_token': self.token
            }
        elif many:
            return {
                'size': len(resource)
            }
        return {}

    def get_fields(self):
        fields = super(UserSerializer, self).get_fields()
        limited_fields = getattr(self.Meta, 'limited_fields', fields)
        request = self.context.get('request', None)
        instance = self.instance
        if (instance and request is not None and request.user != instance) or \
                type(instance) == list:
            result = fields.copy()
            for field in fields:
                if field not in limited_fields:
                    result.pop(field)
            return result
        return fields

    class Meta:
        model = User
        fields = ('url', 'email', 'password', 'old_password', 'groups')


def init_kwargs(model, kwargs):
    return {
        k: v for k, v in kwargs.items() if k in [
            f.name for f in model._meta.get_fields()
        ]
    }



import jwt
import datetime
from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework.authentication import BaseAuthentication
from rest_framework import exceptions, views

User = get_user_model()


def get_token(user):
    token = jwt.encode(
        {
            'exp': datetime.datetime.utcnow() +
            getattr(
                settings,
                'JASONPI_TOKEN_DURATION',
                datetime.timedelta(days=1),
            ),
            'user_id': user.id,
        },
        settings.SECRET_KEY,
        algorithm='HS256'
    )
    if type(token) == bytes:
        return token.decode('utf-8')
    else:
        return token


class JWTAuthentication(BaseAuthentication):
    def authenticate(self, request):
        token = None
        authorization = request.META.get('HTTP_AUTHORIZATION', None)
        if authorization is not None and authorization.find('Bearer ') == 0:
            token = authorization[7:]
        elif 'access_token' in request.COOKIES:
            token = request.COOKIES['access_token']
        if token is None:
            return None
        try:
            return (User.objects.get(
                pk=jwt.decode(
                    token,
                    settings.SECRET_KEY, algorithm='HS256'
                )['user_id']
            ), token)
        except Exception as e:
            raise exceptions.AuthenticationFailed('Invalid token')

    def authenticate_header(self, request):
        return 'Bearer'


def custom_exception_handler(exc, context):
    response = views.exception_handler(exc, context)

    if response is not None and response.status_code == 401:
        # we should delete the access_token cookie
        response.delete_cookie('access_token')

    return response



from django.contrib.auth import get_user_model
from django.db import models


class Provider(models.Model):
    uid = models.CharField(max_length=255)
    provider = models.CharField(max_length=255)
    user = models.ForeignKey(
        get_user_model(),
        on_delete=models.CASCADE,
        related_name='providers',
    )

    class Meta:
        unique_together = (('uid', 'provider'), )



import datetime


def google_profile(data):
    profile = {
        'email': data['emailAddresses'][0]['value'],
        'first_name': data['names'][0]['givenName'],
        'last_name': data['names'][0]['familyName'],
        'gender': data['genders'][0]['value'],
    }
    if 'birthdays' in data and len(data['birthdays']) > 0:
        bd = data['birthdays'][0]['date']
        profile['birthday'] = datetime.date(bd['year'], bd['month'], bd['day'])
    return profile


def facebook_profile(data):
    profile = data.copy()
    try:
        profile['birthday'] = \
            datetime.datetime.strptime(data['birthday'], "%m/%d/%Y").date()
    except ValueError:
        pass
    return profile



from django.contrib.auth.models import (
    BaseUserManager,
    AbstractUser,
)
from django.db import models
from django.utils.translation import ugettext_lazy as _


class UserEmailManager(BaseUserManager):
    """Define a model manager for User model with no username field."""

    use_in_migrations = True

    def _create_user(self, email, password, **extra_fields):
        """Create and save a User with the given email and password."""
        if not email:
            raise ValueError('The given email must be set')
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        """Create and save a regular User with the given email and password."""
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        """Create and save a SuperUser with the given email and password."""
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')
        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')

        return self._create_user(email, password, **extra_fields)


class JasonPIUser(AbstractUser):
    objects = UserEmailManager()

    REQUIRED_FIELDS = []
    USERNAME_FIELD = 'email'

    username = None
    email = models.EmailField(_('email address'), unique=True)
    birthday = models.DateField(null=True, blank=True)
    gender = models.CharField(max_length=255, null=True, blank=True)

    def __str__(self):
        return '%s %s <%s>' % (self.first_name, self.last_name, self.email)

    class Meta:
        abstract = True



from datetime import datetime, date, timezone
import os
import hashlib
import base64
import hmac
import json
import uuid
import boto3
import re

from django.http import HttpResponseBadRequest
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils.module_loading import import_string
from django.contrib.auth import get_user_model
from rest_framework import \
    viewsets, \
    exceptions, \
    parsers, \
    mixins, \
    renderers as default_renderers
from rest_framework.decorators import \
    api_view, \
    permission_classes, \
    renderer_classes
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.permissions import AllowAny
from rest_framework_json_api import renderers, parsers as jsonapi_parsers

from jasonpi.serializers import \
    UserSerializer as JPIUserSerializer, \
    ProviderSerializer
from jasonpi.utils import get_token

User = get_user_model()

if hasattr(settings, 'USER_SERIALIZER'):
    UserSerializer = import_string(settings.USER_SERIALIZER)
else:
    UserSerializer = JPIUserSerializer

if hasattr(settings, 'S3_KEY_GETTER'):
    s3_key_getter = import_string(settings.S3_KEY_GETTER)
else:
    def s3_key_getter(request, ext):
        return 'u%d/misc/%s_%s.%s' % (
            request.user.id,
            date.today().strftime('%Y-%m-%d'),
            uuid.uuid4(),
            ext
        )


class AuthSignInView(APIView):
    parser_classes = (parsers.JSONParser, )
    permission_classes = (AllowAny, )
    renderer_classes = (renderers.JSONRenderer, )
    resource_name = 'users'

    def post(self, request):
        msg = _('Invalid credentials')
        try:
            email = request.data.get('email')
            password = request.data.get('password')
        except AttributeError:
            raise exceptions.ValidationError(msg)
        try:
            user = User.objects.get(email=email)
        except User.DoesNotExist:
            raise exceptions.ValidationError(msg)
        if not user.check_password(password):
            raise exceptions.ValidationError(msg)
        token = get_token(user)
        user.last_login = datetime.now(timezone.utc)
        user.save()
        user_serializer = UserSerializer(user, context={'request': request})
        response = Response(user_serializer.data)
        response.set_cookie(
            'access_token',
            token,
            secure=not settings.DEBUG,
            httponly=True
        )
        user_serializer.token = token
        return response


@api_view()
@permission_classes([AllowAny])
def signout(request):
    response = Response(status=204)
    response.delete_cookie('access_token')
    return response


class AuthProviderView(APIView):
    parser_classes = (parsers.JSONParser, )
    permission_classes = (AllowAny, )
    renderer_classes = (renderers.JSONRenderer, )
    resource_name = 'users'

    def post(self, request):
        try:
            provider_serializer = ProviderSerializer(
                data=request.data,
                context={'request': request}
            )
            provider_serializer.is_valid(raise_exception=True)
            provider_serializer.save()
            user = provider_serializer.instance
            token = get_token(user)
            user.last_login = datetime.now(timezone.utc)
            user.save()
            user_serializer = UserSerializer(
                user,
                context={'request': request},
            )
            response = Response(user_serializer.data)
            response.set_cookie(
                'access_token',
                token,
                secure=not settings.DEBUG,
                httponly=True
            )
            user_serializer.token = token
            return response
        except Exception as e:
            raise e


class AuthRegisterView(mixins.CreateModelMixin,
                       viewsets.GenericViewSet):
    parser_classes = (jsonapi_parsers.JSONParser, )
    serializer_class = UserSerializer
    permission_classes = (AllowAny, )
    resource_name = 'users'

    def create(self, request, *args, **kwargs):
        response = super(AuthRegisterView, self).create(
            request,
            *args,
            **kwargs
        )
        token = get_token(response.data.serializer.instance)
        response.set_cookie(
            'access_token',
            token,
            secure=not settings.DEBUG,
            httponly=True
        )
        response.data.serializer.token = token
        return response


@api_view()
@renderer_classes([default_renderers.JSONRenderer])
def s3_get_presigned_url(request):
    s3 = boto3.client('s3')
    object_name = request.GET.get('objectName')
    ext = object_name.split('.')[-1]
    key = s3_key_getter(request, ext)
    bucket = settings.BUCKET
    user_id = request.user.id
    regex = re.compile(
        r'u?%d/(assets|misc)/[0-9]{4}-[0-9]{2}-[0-9]{2}_[0-9a-z-]+\.[^\.]+' %
        user_id,
    )
    if not regex.match(key):
        raise HttpResponseBadRequest({'error': 'Wrong S3 Key'})
    params = {
        'Bucket': bucket,
        'Key': key,
        'ACL': 'public-read',
    }
    content_type = request.GET.get('contentType', None)
    if content_type is not None:
        params['ContentType'] = content_type
    url = s3.generate_presigned_url(
        'put_object',
        Params=params,
        ExpiresIn=3600
    )
    return Response({'signedUrl': url})


@api_view(['POST'])
@renderer_classes([default_renderers.JSONRenderer])
def s3_sign_policy_document(request):
    request_payload = json.loads(request.body)
    headers = request_payload.get('headers', None)
    if headers:
        # The presence of the 'headers' property in the request payload
        # means this is a request to sign a REST/multipart request
        # and NOT a policy document
        response_data = sign_headers(headers)
    else:
        response_data = sign_policy_document(request_payload)
    return Response(response_data)


def sign_headers(headers):
    """ Sign and return the headers for a chunked upload. """
    return {
        'signature': base64.b64encode(hmac.new(
            os.environ.get('AWS_SECRET_ACCESS_KEY').encode('utf-8'),
            headers.encode('utf-8'),
            hashlib.sha1,
        ).digest()).decode('utf-8'),
    }


def sign_policy_document(policy_document):
    """ Sign and return the policy doucument for a simple upload.
    http://aws.amazon.com/articles/1434/#signyours3postform
    """
    policy = base64.b64encode(json.dumps(policy_document).encode('utf-8'))
    signature = base64.b64encode(hmac.new(
        os.environ.get('AWS_SECRET_ACCESS_KEY').encode('utf-8'),
        policy,
        hashlib.sha1,
    ).digest())
    return {
        'policy': policy.decode('utf-8'),
        'signature': signature.decode('utf-8'),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

setup(
    name='jasonpi',
    version='0.3.0',
    description='Authentication package to use with '
                'django-rest-framework & json-api',
    url='http://github.com/gobadiah/jasonpi',
    author='Michaël Journo',
    author_email='gobadiah@gmail.com',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'pyjwt',
        'httplib2',
        'oauth2client',
        'google-api-python-client',
        'facebook-sdk',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    zip_safe=True,
)



from django.conf.urls import url

from jasonpi import views

urlpatterns = [
    url(r'^auth/s3$', views.s3_get_presigned_url),
    url(r'^auth/s3/signature$', views.s3_sign_policy_document),
    url(r'^auth/signin$', views.AuthSignInView.as_view()),
    url(r'^auth/signout$', views.signout),
    url(
        r'^auth/register$',
        views.AuthRegisterView.as_view({'post': 'create'}),
    ),
    url(r'^auth/provider$', views.AuthProviderView.as_view()),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-24 15:14
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('jasonpi', '0001_initial'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='provider',
            unique_together=set([('uid', 'provider')]),
        ),
    ]
